#!/usr/bin/env python
import argparse
from backports import csv
import io
import isbnlib
import re

fieldnames = ['Number', 'MediaType', 'Title',
              'Authors','Publisher','Year','ISBN',
              'Area','Subject','Language',
              'Source','Acquired',
              'Location',
              'Read','Lent',
              'Comments',
              'webchecked']

def canonicalize_title(raw_title):
    return ' '.join(re.split("[ -:;,/]+", raw_title.lower()))

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--batchsize", "-b", type=int, default=8)
    parser.add_argument("input")
    parser.add_argument("output")
    args = parser.parse_args()
    countdown = args.batchsize
    with io.open(args.input, 'r', encoding='utf-8') as input:
        books_reader = csv.DictReader(input)
        with io.open(args.output, 'w', encoding='utf-8') as output:
            books_writer = csv.DictWriter(output, fieldnames)
            books_writer.writeheader()
            for row in books_reader:
                if countdown > 0 and not row.get('webchecked', None):
                    isbn = str(row.get('ISBN', None))
                    if len(isbn) == 9:
                        isbn = "0" + isbn
                    if isbn:
                        countdown = countdown - 1
                        new_isbn = isbnlib.to_isbn13(isbnlib.canonical(isbn))
                        if new_isbn is None or new_isbn == "":
                            print "Could not canonicalize isbn", isbn
                        else:
                            row['ISBN'] = new_isbn
                        details = None
                        try:
                            details = isbnlib.meta(isbn)
                        except isbnlib.dev._exceptions.NoDataForSelectorError:
                            print "No data for ISBN", isbn, "title", row.get('Title', "Unknown")
                            row['webchecked'] = "No data for ISBN"
                        except isbnlib._exceptions.NotValidISBNError:
                            print "Invalid ISBN", isbn, "for", row['Title']
                            row['webchecked'] = "Invalid ISBN"
                        except isbnlib.dev._exceptions.ISBNNotConsistentError:
                            print "Inconsistent data for",  row['Title']
                            row['webchecked'] = "Inconsistent ISBN data"
                        if details:
                            if details.get('ISBN-13', "") != "" and row.get('ISBN', "") == "":
                                row['ISBN'] = details['ISBN-13']
                            if 'Authors' in row:
                                row['Authors'] = row['Authors'].split('/')
                            old_title = row['Title']
                            web_title = details['Title']
                            if old_title != web_title:
                                old_canon = canonicalize_title(old_title)
                                web_canon = canonicalize_title(web_title)
                                old_len = len(old_canon)
                                web_len = len(web_canon)
                                if ((web_len > old_len and old_canon in web_canon)
                                    or (web_len == old_len and old_canon == web_canon)):
                                    print "Title improvement from", old_title, "to", web_title
                                else:
                                    print "Title discrepancy:", old_title, "in file,", web_title, "found online"
                                    details['Title'] = old_title
                            # don't use 'update', because we don't want to drag in random other fields that dictwriter will then object to
                            for key in fieldnames:
                                if key in details:
                                    row[key] = details[key]
                            if 'Authors' in row:
                                row['Authors'] = '/'.join(row['Authors'])
                            row['webchecked'] = "OK"
                # from https://docs.python.org/2/library/csv.html
                encoded_row = {k: (v.encode("utf-8") if isinstance(v, basestring) else v)
                               for k,v in row.iteritems()}
                books_writer.writerow(row)

if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
from backports import csv
import io
import sys
import os

# Fields to split into lists
multi_fields = ['Parents', 'Offspring', 'Siblings',
                'Partners', 'Ex-partners',
                'Knows',
                'Organizations']

def make_name(person):
    return ' '.join([person.get('Given name', "")]
                    + person.get('Middle names', "").split(" ")
                    + [person.get('Surname', "")])

def make_name_list(people):
    # todo: collect together all the people of the same surname and format that as "Fred, Freddie and Freda Smith and John and Jane Jones"
    return ", ".join([resi['_name_'] for resi in people])

def assemble_postal_address(who, spacer="; "):
    """Return a postal address."""
    address = []
    for key in ['Street', 'City', 'Region', 'Postal Code', 'Country']:
        value = who[key]
        if value != "":
            address += value.split("; ")
    return spacer.join(address)

def show_person(out, who):
    out.write(who['_name_'] + "\n")
    email = who['Primary email']
    if email != "":
        out.write("    Email: " + email + "\n")
    phone = who['Primary phone Value']
    if phone != "":
        phone_type = who['Primary phone Type']
        out.write("    " + (phone_type if phone_type != "" else "Phone") + ': ' + phone + "\n")
    if who['Street'] != "":
        out.write("    Address:\n        "
                  + assemble_postal_address(who, "\n        ")
                  + "\n")

def main():
    parser = argparse.ArgumentParser()
    org_files = os.environ.get("ORG", "~/org")
    parser.add_argument("--contacts", "-c",
                        default=os.path.join(org_files, "contacts.csv"),
                        help="""The CSV file containing the contact details.""")
    parser.add_argument("--flag", "-f",
                        action='store_true',
                        help="""Search by flag.""")
    parser.add_argument("names",
                        nargs='*',
                        help="""The names to look for.""")
    args = parser.parse_args()
    by_name = {}
    by_id = {}
    with io.open(args.contacts, 'r', encoding='utf-8') as input:
        contacts_reader = csv.DictReader(input)
        for row in contacts_reader:
            for multi in multi_fields:
                row[multi] = (row.get(multi, "") or "").split(";")
            n = make_name(row)
            row['_name_'] = n
            by_name[n] = row
            by_id[id] = row.get('ID', "")
    if args.flag != "":
        flag = args.names[0]
        by_address = {}
        for n in sorted(by_name.keys()):
            who = by_name[n]
            if flag in who['Flags']:
                addr = assemble_postal_address(who, "\n  ")
                if addr not in by_address:
                    by_address[addr] = []
                by_address[addr].append(who)
        for addr in sorted(by_address.keys()):
            sys.stdout.write(make_name_list(by_address[addr])
                             + "\n  " + addr + "\n\n")
    else:
        name = " ".join(args.names)
        if name in by_name:
            show_person(sys.stdout, by_name[name])
        else:
            for n in sorted(by_name.keys()):
                if name in n:
                    show_person(sys.stdout, by_name[n])

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
import client_server # the shell script ./storage-client makes this available
import decouple

def main():
    parser=argparse.ArgumentParser()
    parser.add_argument('data', nargs='*', action='append',
                        help="""The data to send to the server.""")
    client_server.client_server_add_arguments(parser, 9797)
    args=parser.parse_args()
    args.server = False
    query_passphrase = decouple.config('query_passphrase')
    reply_passphrase = decouple.config('reply_passphrase')
    client_server.check_private_key_privacy(args)
    query_key, reply_key = client_server.read_keys_from_files(args,
                                                              query_passphrase,
                                                              reply_passphrase)
    text = " ".join(args.data[0])

    received = client_server.get_response(
        text,
        args.host, args.port, args.tcp,
        encryption_scheme=ord('H' if query_key and reply_key else 'p'),
        representation_scheme=ord('p'),
        query_key=query_key,
        reply_key=reply_key)

    print(received)
    return 0

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import re
import json
import sys

import coimealta.contacts.contacts_data as contacts_data

def safesearch(flags, text):
    print("Searching", text)
    return flags.search(text)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--flag", action='append')
    parser.add_argument("-g", "--group", action='append')
    parser.add_argument("-a", "--all", action='store_true')
    parser.add_argument("-s", "--surname", action='append')
    parser.add_argument("-G", "--given", action='append')
    parser.add_argument("-N", "--no-add-family",
                        help="""Without this option, if someone is selected but their partner
                        or children aren't, those are added automatically to the selection.""")
    parser.add_argument("-p", "--postal-addresses",
                        action='store_true',
                        help="""List people by address, grouping together those at the same address.
                        Without this, people are listed individually, with their email addresses.""")
    parser.add_argument("-e", "--email-addresses",
                        action='store_true',
                        help="""List people by email address.""")
    parser.add_argument("-j", "--json", action='store_true',
                        help="""Output full details as JSON.""")
    parser.add_argument("input")
    args = parser.parse_args()
    by_id, _ = contacts_data.read_contacts(args.input)
    if args.all:
        selected = by_id.values()
    else:
        selected = []
        if args.flag:
            flags = re.compile("[" + "".join(args.flag) + "]")
            selected += [someone for someone in by_id.values()
                         if flags.search(someone.get('Flags', "") or "")]
        if args.group:
            groups = set(args.group)
            selected += [someone for someone in by_id.values()
                         if len((groups.intersection(someone['_groups_']))) > 0]
        if args.surname:
            surnames = set(args.surname)
            selected += [someone for someone in by_id.values()
                         if someone['Surname'] in surnames]
        if args.given:
            given_names = set(args.given)
            selected += [someone for someone in by_id.values()
                         if someone['Given name'] in given_names]
        if not args.no_add_family:
            invited_ids = [whoever['ID'] for whoever in selected]
            for whoever in selected:
                for partner in whoever['Partners']:
                    if partner not in invited_ids:
                        # todo: make this only if they are at the same address
                        selected.append(by_id[partner])
            for whoever in selected:
                for offspring in whoever['Offspring']:
                    if offspring not in invited_ids:
                        print(contacts_data.make_name(by_id[offspring]), "  -- maybe add")
                        # todo: make this only if they are at the same address
                        selected.append(by_id[offspring])
    if args.postal_addresses:
        by_address = {}
        for contact in selected:
            address = contacts_data.make_address(contact)
            if address in by_address:
                by_address[address].append(contact)
            else:
                by_address[address] = [contact]
        print(len(by_address), "addresses")
        print("")
        for addr, residents in by_address.items():
            print(contacts_data.names_string(residents))
            print("  " + "\n  ".join([a for a in addr if a != ""]))
            print("")
    else:
        for contact in selected:
            if args.email_addresses:
                email = contact['Primary email']
                if email != "":
                    print(email + " <" + contact['_name_'] + ">")
            elif args.json:
                json.dump(contact, sys.stdout)
            else:
                print(contact['_name_'])

if __name__ == "__main__":
    main()



#!/usr/bin/python3

# import argparse
import csv
import os
import datetime

PERISHABLES_FILE = os.path.expandvars("$SYNCED/org/perishables.csv")

def convert_row(row):
    row['Best before'] = datetime.date.fromisoformat(row['Best before'])
    row['Quantity'] = 1 if row.get('Quantity', "") == "" else int(row.get('Quantity', "1"))
    return row

def get_perishables():
    with open(PERISHABLES_FILE) as instream:
        return sorted([convert_row(raw) for raw in csv.DictReader(instream)],
                      key=lambda row: row['Best before'])

def perishables_main():
    for row in get_perishables():
        print(row['Best before'], row['Product'], "x", row['Quantity'])

def main():
    # parser = argparse.ArgumentParser()
    # args = parser.parse_args()
    perishables_main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import argparse
import cmd
import collections
import csv
import decouple
import functools
import json
import math
import operator
import os
import re
import shlex
import sys
# import yaml

STORAGE_BASE=500000

HAS_CLIENT_SERVER = True
try:
    import simple_client_server.client_server as client_server
except:
    HAS_CLIENT_SERVER = False

class Storer:

    def __init__(self, locations, items, books, initial_type="books"):
        self.locations = locations
        self.items = items
        self.books = books
        self.current_type = initial_type
        self.current_location = None

    def store(self, token):
        if token in ('books', 'items'):
            self.current_type = token
        elif token == 'quit':
            return False
        else:
            token = int(token)
            if token >= STORAGE_BASE:
                self.current_location = token
            else:
                if self.current_type == 'book':
                    store_book(self.books, token, self.current_location)
                else:
                    store_item(self.items, token, self.current_location)
            return True

class StorageShell(cmd.Cmd):

    prompt = "Storage> "

    def __init__(self, outstream, locations, items, books):
        super().__init__()
        self.outstream = outstream
        self.locations = locations
        self.items = items
        self.books = books

    def postcmd(self, stop, _line):
        return stop

    def do_list_books(self, *_args):
        """Show a table of where all the books are."""
        by_location = collections.defaultdict(list)
        for idx, book in self.books.items():
            if 'Location' in book:
                by_location[book['Location']].append(idx)
        for loc in sorted(by_location.keys()):
            contents = by_location[loc]
            self.outstream.write(describe_nested_location(self.locations, loc) + ":\n")
            for title in sorted([self.books[idx]['Title'] for idx in contents ]):
                self.outstream.write("    " + title + "\n")
        return False

    def do_capacities(self, *_args):
        """Analyze the storage capacities.
    Shows how much of each type of storage there is, and also a summary
    combining the types."""
        capacity_by_type, volume, bookshelf_length, other_length, area = calculate_capacities(self.locations)
        for loctype in sorted(capacity_by_type.keys()):
            label_width = max(*map(len, capacity_by_type.keys()))
        self.outstream.write(loctype.rjust(label_width) + " " + str(math.ceil(capacity_by_type[loctype])) + "\n")
        self.outstream.write("Total container volume: " + str(volume) + " litres\n")
        self.outstream.write("Total container and book (estimate) volume: "
                             + str(volume + bookshelf_length * bookshelf_area)
                             + " litres\n")
        self.outstream.write("Total shelving length: "
                             + str(bookshelf_length + other_length)
                             + " metres\n")
        self.outstream.write("Total panel area: "
                             + str(area)
                             + " square metres\n")
        return False

    def do_counts(self, *_args):
        """Count how many of each type of thing I have."""
        # todo: maybe do the books here as well?
        types = {}
        for item in self.items.values():
            item_type = item['Type']
            if item_type not in types:
                types[item_type] = {}
            subtypes = types[item_type]
            item_subtype = item['Subtype']
            if item_subtype in subtypes:
                subtypes[item_subtype] += 1
            else:
                subtypes[item_subtype] = 1
        for item_type in sorted(types.keys()):
            subtypes = types[item_type]
            self.outstream.write((item_type if item_type != "" else "unspecified")
                            + ": "
                            + str(functools.reduce(operator.add, subtypes.values()))
                            + "\n")
            for subtype in sorted([k for k in subtypes.keys() if k is not None ]):
                self.outstream.write("    "
                                     + (subtype
                                        if subtype != "" and subtype is not None
                                        else "unspecified")
                                     + ": "
                                     + str(subtypes[subtype])
                                     + "\n")
        return False

    def do_list_items(self, *args):
        """Show a table of where all the items are."""
        # TODO: option to print table of where all inventory items are
        by_location = defaultdict(list)
        for idx, item in self.items.items():
            if 'Normal location' in item:
                by_location[item['Normal location']].append(idx)
        for loc in sorted(by_location.keys()):
            contents = by_location[loc]
            self.outstream.write(describe_nested_location(self.locations, loc) + ":\n")
            for title in sorted([self.items[idx]['Item'] for idx in contents ]):
                self.outstream.write("    " + title + "\n")
        return False

    def do_name_completions(self, *things):
        """Return the names matching a fragment."""
        if len(things) == 0:
            return ""
        fragment = things[0]
        self.outstream.write(json.dumps(sorted(
            [book['Title']
              for book in self.books.values()
              if fragment in book['Title']]
            + [item['Item']
               for item in self.items.values()
               if fragment in item['Item']]))
                             + "\n")

    def do_location_completions(self, *things):
        """Return the location names matching a fragment."""
        if len(things) == 0:
            return ""
        fragment = things[0]
        self.outstream.write(json.dumps(sorted(
            [location['Description']
             for location in self.locations.values()
             if fragment in location['Description']]))
            + "\n")
        return False

    def do_quit(self, _args):
        """Stop the CLI."""
        return True

    def do_list_locations(self, *things):
        """List everything that is in the matching locations."""
        for where in sorted(locations_matching_patterns(self.locations, things)):
            list_location(self.outstream, where, "", self.locations, self.items, self.books)
        return False

    def do_find_things(self, *args):
        """Show the locations of things.
        This finds books, other items, and locations."""
        print("Looking for", *args)
        findings = {}
        for thing in args:
            if re.match("[0-9]+", thing):
                as_location = describe_nested_location(self.locations, thing)
                if as_location != []:
                    findings[thing] = as_location
            for book in books_matching(self.books, thing):
                findings[book['Title']] = describe_nested_location(self.locations, book['Location'])
            for item in items_matching(self.items, thing):
                findings[item['Item']] = describe_nested_location(self.locations, item['Normal location'])
        for finding in sorted(findings.keys()):
            self.outstream.write(finding + " is " + findings[finding] + "\n")
        return False

    def do_store(self, thing_type="books", *args):
        """Put things into locations."""
        storer = Storer(self.locations, self.items, self.books, initial_type=thing_type)
        if args:
            for arg in args:
                storer.store(arg)
        else:
            done = False
            while not done:
                for token in sys.stdin.readline().split():
                    if not storer.store(token):
                        done = True
                        break

def normalize_book_entry(row):
    ex_libris = row['Number']
    if isinstance(ex_libris, str) and ex_libris != "":
        ex_libris = int(ex_libris)
        row['Number'] = ex_libris
    location = row['Location']
    if isinstance(location, str) and location != "":
        location = int(location)
        row['Location'] = location
    return row

def read_books(books_file, _key=None):
    with open(os.path.expandvars(books_file)) as instream:
        return { book['Number']: book
                 for book in [normalize_book_entry(row)
                              for row in csv.DictReader(instream)]
                 if book['Number'] != "" }

# Description for reading these files using client_server.py:
# ('Number', normalize_book_entry)

def book_matches(book, pattern):
    pattern = re.compile(pattern , re.IGNORECASE)
    return (book['Title'] and pattern.search(book['Title'])
            or book['Authors'] and pattern.search(book['Authors'])
            or book['Publisher'] and pattern.search(book['Publisher'])
            or book['ISBN'] and pattern.search(book['ISBN'])
            or book['Area'] and pattern.search(book['Area']))

def books_matching(book_index, pattern):
    return [book
            for book in book_index.values()
            if book_matches(book, pattern) ]

unlabelled = -1

def normalize_item_entry(row):
    global unlabelled
    label_number = row.get('Label number', "")
    if isinstance(label_number, str) and label_number != "":
        label_number = int(label_number)
        row['Label number'] = label_number
    else:
        row['Label number'] = unlabelled
        unlabelled -= 1
    normal_location = row['Normal location']
    if isinstance(normal_location, str) and re.match("[0-9]+", normal_location):
        normal_location = int(normal_location)
        row['Normal location'] = normal_location
    return row

def read_inventory(inventory_file, _key=None):
    if os.path.exists(inventory_file):
        with open(os.path.expandvars(inventory_file)) as instream:
            return {item['Label number']: item
                    for item in map(normalize_item_entry,
                                    [row
                                     for row in csv.DictReader(instream)])}
    else:
        return {}

# Description for reading these files using client_server.py:
# ('Label number', normalize_item_entry)

def item_matches(item, pattern):
    pattern = re.compile(pattern , re.IGNORECASE)
    return (pattern.search(item['Item'])
            or (item['Type'] and pattern.search(item['Type']))
            or (item['Subtype'] and pattern.search(item['Subtype'])))

def items_matching(inventory_index, pattern):
    return [item
             for item in inventory_index.values()
             if item_matches(item, pattern) ]

def store_item(inventory_index, item, location):
    """Record that an item is in a location."""
    inventory_index[item]['Normal location'] = location

def store_book(inventory_index, book, location):
    """Record that a book is in a location."""
    inventory_index[book]['Location'] = location

def normalize_location(row):
    contained_within = row['ContainedWithin']
    try:
        row['ContainedWithin'] = (int(contained_within)
                                  if (isinstance(contained_within, str)
                                      and contained_within != "")
                                  else None)
    except ValueError:
        row['ContainedWithin'] = None
    number = row['Number']
    row['Number'] = int(number) if (isinstance(number, str)
                                    and number != "") else None
    return row

def read_locations(locations_file, _key=None):
    with open(os.path.expandvars(locations_file)) as instream:
        return { location['Number']: location
                 for location in [normalize_location(row)
                                   for row in csv.DictReader(instream)
                                   if row['Number'] is not None ] }

# Description for reading these files using client_server.py:
# ('Number', normalize_location)

def locations_matching(locations_index, pattern):
    """Return a list of location numbers for locations that match a regexp."""
    if pattern == "all":
        return [loc['Number']
                 for loc in locations_index.values() ]
    else:
        pattern = re.compile(pattern, re.IGNORECASE)
        return [loc['Number']
                 for loc in locations_index.values()
                 if pattern.search(loc['Description']) ]

def locations_matching_patterns(locations_index, patterns):
    """Return a set of location numbers for locations that match any of a list of regexps."""
    result = set()
    for pattern in patterns:
        for loc in locations_matching(locations_index, pattern):
            result.add(loc)
    return result

def describe_location(where):
    """Return a description of a location."""
    description = where['Description'].lower()
    level = where['Level']
    if level != "":
        description += " "
        if re.match("[0-9]+", level):
            description += "level " + level
        else:
            description += level
    storage_type = where['Type']
    if storage_type != "":
        if storage_type in ("room", "building"):
            description = "the " + description
        else:
            if (storage_type == "shelf"
                and not re.search("shelves", description)):
                description += " " + storage_type
    description = ("on " if storage_type == "shelf" else "in ") + description
    return description

def nested_location(locations, location):
    result = []
    try:
        location = int(location)
        while location:
            if location not in locations:
                break
            where = locations[location]
            description = describe_location(where)
            result.append(description)
            location = where['ContainedWithin']
        return result
    except:
        return ["Could not follow location %s" % location]

def describe_nested_location(locations, location):
    """Return a description of a location, along with any surrounding locations."""
    return (" which is ".join(nested_location(locations, location))
            if location != ""
            else "unknown")

def sum_capacities(all_data, types):
    return math.ceil(functools.reduce(operator.add,
                                      [all_data.get(loctype, 0)
                                       for loctype in types ]))

# factor to convert meter of bookshelf to litre of books
# the 10 is because a liter is a decimeter along each side
bookshelf_area = 10 * 2 * 1.5

def calculate_capacities(locations):
    capacity_by_type = {}
    for location in locations.values():
        loctype = (location['Type'] or "").lower() if 'Type' in location else ""
        locsize = (location['Size'] or "").lower() if 'Size' in location else ""
        if loctype != "" and locsize != "":
            capacity_by_type[loctype] = float(capacity_by_type.get(loctype, 0)) + float(locsize)
    volume = sum_capacities(capacity_by_type, ('box', 'crate',
                                               'drawer', 'cupboard'))
    bookshelf_length = sum_capacities(capacity_by_type, ('bookshelf', 'bookshelves'))
    other_length = sum_capacities(capacity_by_type, ('shelf', 'shelves',
                                                     'cupboard shelf', 'racklevel'))
    area = sum_capacities(capacity_by_type, ('louvre panel', 'pegboard'))
    return capacity_by_type, volume, bookshelf_length, other_length, area

def list_location(outstream, location, prefix, locations, items, books):
    """List everything that is in the given location."""
    if type(location) == dict:
        location = location['Number']
    directly_contained_items = [
        item for item in items.values()
        if item['Normal location'] == location ]
    directly_contained_books = [
        book for book in books.values()
        if book['Location'] == location ]
    sub_locations = [
        subloc for subloc in locations.values()
        if subloc['ContainedWithin'] == location ]
    description = describe_location(locations[location])
    next_prefix = prefix + "    "
    if len(directly_contained_items) > 0:
        outstream.write(prefix + "Items directly " + description + ":\n")
        for item in directly_contained_items:
            outstream.write(next_prefix + item['Item'] + "\n")
    if len(directly_contained_books) > 0:
        outstream.write(prefix + "Books directly " + description + ":\n")
        for book in directly_contained_books:
            outstream.write(next_prefix + book['Title'] + "\n")
    if len(sub_locations) > 0:
        outstream.write(prefix + "Locations " + description + ":\n")
        for subloc in sub_locations:
            outstream.write(next_prefix + subloc['Description'] + "\n")
            list_location(outstream, subloc, next_prefix, locations, items, books)

filenames = {}

remembered_items_data = {'combined': None,
                         'inventory': None,
                         'stock': None,
                         'project_parts': None,
                         'books': None,
                         'locations': None}

def storage_server_function(in_string, files_data):
    command_parts = shlex.split(in_string)
    if len(command_parts) > 0:
        inventory = files_data[filenames['inventory']]
        stock = files_data[filenames['stock']]
        project_parts = files_data[filenames['project_parts']]
        if (inventory is not remembered_items_data['inventory']
            or stock is not remembered_items_data['stock']
            or project_parts is not remembered_items_data['project_parts']):
            items_data = inventory
            items_data.update(stock)
            items_data.update(project_parts)
            remembered_items_data['combined'] = items_data
            remembered_items_data['inventory'] = inventory
            remembered_items_data['stock'] = stock
            remembered_items_data['project_parts'] = project_parts
        else:
            items_data = remembered_items_data['combined']
        output_catcher = io.StringIO()
        StorageShell(output_catcher,
                     files_data[filenames['locations']],
                     items_data,
                     files_data[filenames['books']]).onecmd(in_string)
        return output_catcher.getvalue()
    else:
        return "Command was empty"

def get_args():
    parser = argparse.ArgumentParser()
    # parser.add_argument("--config", "-c",
    #                     default="/usr/local/share/storage.yaml",
    #                     help="""The config file for the storage system.""")
    parser.add_argument("--locations", "-f",
                        default=os.path.expandvars("$ORG/storage.csv"),
                        help="""The CSV file containing the storage locations.""")
    parser.add_argument("--books", "-b",
                        default=os.path.expandvars("$ORG/books.csv"),
                        help="""The CSV file containing the book catalogue.""")
    parser.add_argument("--inventory", "-i",
                        default=os.path.expandvars("$ORG/inventory.csv"),
                        help="""The CSV file containing the general inventory.""")
    parser.add_argument("--stock", "-s",
                        default=os.path.expandvars("$ORG/stock.csv"),
                        help="""The CSV file containing the stock material inventory.""")
    parser.add_argument("--project-parts", "-p",
                        default=os.path.expandvars("$ORG/project-parts.csv"),
                        help="""The CSV file containing the project parts inventory.""")
    actions = parser.add_mutually_exclusive_group()
    actions.add_argument("--server", action='store_true',
                        help="""Run a little CLI on a network socket.""")
    actions.add_argument("--cli", action='store_true',
                         help="""Run a little CLI on stdin and stdout.""")
    if HAS_CLIENT_SERVER:
        client_server.client_server_add_arguments(parser, 9797, include_keys=False)
    parser.add_argument("things",
                        nargs='*',
                        help="""The things to look for.""")
    return vars(parser.parse_args())

def storage(locations,
            books,
            inventory,
            stock,
            project_parts,
            server: bool,
            cli: bool,
            host: str,
            port,
            tcp: bool,
            things):
    if server:
        global filenames
        filenames = {'inventory': os.path.basename(inventory),
                     'books': os.path.basename(books),
                     'stock': os.path.basename(stock),
                     'project_parts': os.path.basename(project_parts),
                     'locations': os.path.basename(locations)}
        if HAS_CLIENT_SERVER:
            query_passphrase = decouple.config('query_passphrase')
            reply_passphrase = decouple.config('reply_passphrase')
            client_server.check_private_key_privacy(args)
            query_key, reply_key = client_server.read_keys_from_files(args,
                                                                      query_passphrase,
                                                                      reply_passphrase)
            client_server.run_servers(host, int(port),
                                      getter=storage_server_function,
                                      files={inventory: ('Label number',
                                                              normalize_item_entry),
                                             books: ('Number',
                                                          normalize_book_entry),
                                             stock: ('Label number',
                                                          normalize_item_entry),
                                             project_parts: ('Label number',
                                                                  normalize_item_entry),
                                             locations: ('Number',
                                                              normalize_location)},
                                      query_key=query_key,
                                      reply_key=reply_key)
    else:
        locations = read_locations(locations)
        items = read_inventory(inventory)
        items.update(read_inventory(stock))
        items.update(read_inventory(project_parts))
        books = read_books(books)
        command_handler = StorageShell(sys.stdout, locations, items, books)
        if cli:
            command_handler.cmdloop()
        else:
            if things[0] in command_handler.completenames(""):
                command_handler.onecmd(" ".join(things))
            else:
                command_handler.onecmd("find_things " + " ".join(things))

if __name__ == "__main__":
    storage(**get_args())



import csv
import datetime
import functools
import operator
import os
import random
import re

def count_grouped_titles(title_map, titles):
    """Count people with titles in a list of titles."""
    return functools.reduce(operator.add,
                            [len(title_map[title]) for title in titles])

fieldnames = ['Given name', 'Middle names', 'Surname', 'Title', 'Old name', 'AKA',
              'Birthday', 'Died',
              'First contact', 'In touch',
              'Gender',
              'ID', 'Parents', 'Offspring', 'Siblings',
              'Partners', 'Ex-partners', 'Knows',
              'Notes',
              'Flags',
              'Nationality', 'Place met',
              'Group Membership', 'Other groups', 'Organizations',
              'Universities', 'Subjects', 'Jobs',
              'Primary email', 'Other emails',
              'Primary phone Type', 'Primary phone Value',
              'Secondary phone Type', 'Secondary phone Value',
              'Street', 'Village/District', 'City', 'County', 'State',
              'Postal Code', 'Country', 'Extended Address']

# Fields to split into lists
multi_fields = ['Parents', 'Offspring', 'Siblings',
                'Partners', 'Ex-partners',
                'Knows',
                'Group Membership',
                'Organizations',
                'Other groups']

def make_name(person):
    """Assemble a name from a person's fields."""
    first_name = (person.get('Given name', "") or "")
    middle_names = [name
                    for name in (person.get('Middle names', "") or "").split(' ')
                    if name]
    surname = (person.get('Surname', "") or "")
    return (first_name + " " + " ".join(middle_names) + " " + surname
            if middle_names
            else first_name + " " + surname)

def make_short_name(person):
    """Make a short form of a person's name."""
    return ' '.join([(person.get('Given name', "") or "")]
                    + [(person.get('Surname', "") or "")])

def string_list_with_and(items):
    return (", ".join(items[:-1])
            + ", and "
            + items[-1] if len(items) > 2 else items[0]
            + " and "
            + items[1] if len(items) == 2 else items[0])

# todo: sort residents to bring oldest (or most ancestral) to the start
def names_string(people):
    by_surname = {}
    for person in people:
        surname = person.get('Surname', "")
        if surname not in by_surname:
            by_surname[surname] = []
        by_surname[surname].append(person.get('Given name', ""))
    names = [string_list_with_and(sorted(by_surname[surname]))
             + " "
             + surname for surname in sorted(by_surname.keys())]
    return string_list_with_and(names)

def make_address(person):
    """Put together an address tuple."""
    return (person.get('Street', ""),
            person.get('Village/District', ""),
            person.get('City', ""),
            person.get('County', ""),
            person.get('State', ""),
            person.get('Postal Code', ""),
            person.get('Country'))

def make_ID():
    """Make a letter-digit-letter-digit random ID that is not a valid hex string."""
    return (str(chr(random.randint(0, 19) + ord('G')))
            + str(chr(random.randint(0, 9) + ord('0')))
            + str(chr(random.randint(0, 25) + ord('A')))
            + str(chr(random.randint(0, 9) + ord('0'))))

def age_in_year(person, year):
    """Return how old a person is in a given year."""
    bday = person.get('Birthday', "") or ""
    match = re.match("[0-9][0-9][0-9][0-9]", bday)
    return (year - int(match.group(0))) if match else None

def birthday(person, this_year):
    """Return this year's birthday of a person, as a datetime.date."""
    bday_string = person.get('Birthday', "") or ""
    if bday_string == "":
        return False
    try:
        bday = datetime.date.fromisoformat(bday_string)
    except Exception:
        match = re.search("-([0-9][0-9])-([0-9][0-9])", bday_string)
        if match:
            month = int(match.group(1))
            if month == 0:
                return False
            day = int(match.group(2))
            if day == 0:
                return False
            bday = datetime.date(year=this_year, month=month, day=day)
        else:
            return False
    bday = bday.replace(year=this_year)
    return bday

def birthday_soon(person, this_year, today, within_days=30):
    """Return whether a person has a birthday soon."""
    bday = birthday(person, this_year)
    if not bday:
        return False
    interval_to_birthday = (bday - today).days
    return interval_to_birthday >= 0 and interval_to_birthday < within_days

def last_contacted(person):
    """Return when I was last in touch with a person, or None if not recorded.
    The input is a person dictionary, and the result is a datetime.date."""
    cday_string = person.get('In touch', "")
    if not cday_string:
        return None
    try:
        cday = datetime.date.fromisoformat(cday_string)
    except:
        match = re.search("([0-9][0-9][0-9][0-9])-([0-9][0-9])", cday_string)
        if match:
            year = int(match.group(1))
            if year == 0:
                return False
            month = int(match.group(2))
            if month == 0:
                month = 1
            cday = datetime.date(year=year, month=month, day=1)
        else:
            match = re.search("([0-9][0-9][0-9][0-9])", cday_string)
            if match:
                cday = datetime.date(year=int(match.group(1)), month=1, day=1)
            else:
                return None
    return cday

def contact_soon(person, today, days_since_last_contact=90):
    """Return whether I haven't registered a contact with someone in a given time."""
    cday = last_contacted(person)
    # TODO: have a contact frequency field in the data for each person
    return cday and (today - cday).days > days_since_last_contact

def record_contact(person, date=None):
    """Record that I have contacted someone on a given date."""
    if date is None:
        date = datetime.date.today()
    set_field_if_greater(person, "last-contacted", date)
    set_field_if_not_empty(person, "keep-in-touch", date)

def age_string(person, year):
    age = age_in_year(person, year)
    return str(age) if age else "?"

def by_name(people, name):
    return people.get(name)

def set_field_if_empty(person, field, value):
    """Set a field of a person dictionary only if there is no value already in that field."""
    if person and (field not in person or person[field] == ""):
        person[field] = value
        return True
    return False

def set_field_if_not_empty(person, field, value):
    """Set a field of a person dictionary only if there is already a value in that field."""
    if person and (field in person and person[field] != ""):
        person[field] = value
        return True
    return False

def set_field_if_greater(person, field, value):
    """Set a field of a person dictionary if the new value is greater than the old one."""
    if (person
        and (field not in person
             or person[field] == ""
             or value > person[field])):
        person[field] = value
        return True
    return False

def set_field(person, field, value):
    """Set a field of a person dictionary.
    None-safe."""
    if person:
        person[field] = value
        return True
    return False

def read_contacts(filename):
    """Read a contacts file and return a tuple of dictionaries.
    The first one lists contacts by ID, and the second by name."""
    people_by_id = {}
    people_by_name = {}
    without_id = []
    with open(os.path.expandvars(filename)) as instream:
        for row in csv.DictReader(instream):
            name = make_name(row)
            short_name = make_short_name(row)
            row['_name_'] = name
            people_by_name[name] = row
            # if short_name != name:
            #     people_by_name[short_name] = row
            uid = row.get('ID', "")
            if uid is not None and uid != "":
                people_by_id[uid] = row
            else:
                without_id.append(row)
            for multi in multi_fields:
                row[multi] = set(
                    item.strip()
                    for item in
                    (row.get(multi, "") or "").split(';')
                    if item
                )
            row['_groups_'] = row['Group Membership'].union(row['Organizations'],
                                                            row['Other groups'])

    for person in without_id:
        uid = make_ID()
        while uid in people_by_id:
            uid = make_ID()
        person['ID'] = uid
        people_by_id[uid] = person

    return people_by_id, people_by_name

def write_contacts(filename, people_by_name):
    """Write a dictionary of contacts-by-name to a file."""
    all_found_fields = set().union(*[set(row.keys()) for row in people_by_name.values()])
    if all_found_fields != set(fieldnames):
        print("These extra fields were found:", all_found_fields - set(fieldnames))
    with open(os.path.expandvars(filename), 'w') as output:
        contacts_writer = csv.DictWriter(output, fieldnames)
        contacts_writer.writeheader()
        for name in sorted(people_by_name.keys()):
            row = people_by_name[name]
            # print row
            for multi in multi_fields:
                # print("converting", multi, row[multi])
                row[multi] = '; '.join(sorted(list(row[multi])))
                # print("converted", multi, row[multi])
            for deledend in ('', '_name_', '_groups_'):
                if deledend in row:
                    del row[deledend]
            try:
                contacts_writer.writerow(row)
            except ValueError:
                print("Unwritable row: ", row)



#!/usr/bin/python3

import argparse

def read_labels(input_file):
    with open(input_file) as instream:
        all_lines = instream.readlines()
        labels = []
        lines = []
        for line in all_lines:
            line = line.strip()
            if line == "":
                labels.append(lines)
                lines = []
            else:
                lines.append(line)
        labels.append(lines)
    return labels

def labels_to_pages(labels, n_per_page):
    return [labels[i:i + n_per_page] for i in range(0, len(labels), n_per_page)]
    
def begin_page(outstream, w, h):
    outstream.write('<svg width="%gmm" height="%gmm" viewBox="0 0 %g %g">\n' % (w, h, w, h))

def end_page(outstream):
    outstream.write('</svg>\n')

def write_label_at(x, y, config, label_lines, outstream):
    leading = config['leading']
    fontsize = config['label-height'] / ((len(label_lines) + 1) * leading)
    if config['verbose']:
        print("placing", label_lines[0], "at", x, y, "using fontsize", fontsize)
    outstream.write('  <g transform="translate(%g,%g)">\n' % (x, y))
    if config['outline']:
        outstream.write('    <rect x="0" y="0" width="%g" height="%g" fill="white" stroke="green"/>\n' % (config['label-width'], config['label-height']))
    outstream.write('    <g transform="translate(%g,%g)">\n' % (config['label-left-margin'], config['label-top-margin']))
    outstream.write('      <text font-size="%g">\n' % fontsize)
    for iline, line in enumerate(label_lines):
        outstream.write('        <tspan x="0" y="%gem">%s</tspan>\n' % ((iline+1) * leading, line))
    outstream.write('      </text>\n')
    outstream.write('    </g>\n')
    outstream.write('  </g>\n')

def main():
    parser = argparse.ArgumentParser()
    # the defaults are for labelplanet's LP12/99 label sheet
    parser.add_argument("-x", "--across",
                        type=int, default=2,
                        help="""How many labels fit across the page.""")
    parser.add_argument("-y", "--down",
                        type=int, default=6,
                        help="""How many labels fit down the page.""")
    parser.add_argument("-H", "--height",
                        type=float, default=42.3,
                        help="""The height of each label (excluding gap, margins etc)""")
    parser.add_argument("-W", "--width",
                        type=float, default=99.1,
                        help="""The width of each label (excluding gap, margins etc)""")
    parser.add_argument("-t", "--top-margin", "--top",
                        type=float, default=21.6,
                        help="""The top margin of the page.""")
    parser.add_argument("-l", "--left-margin", "--left",
                        type=float, default=4.65,
                        help="""The left margin of the page.""")
    parser.add_argument("-g", "--horizontal-gap",
                        type=float, default=2.5,
                        help="""The horizontal gap between labels.""")
    parser.add_argument("-V", "--vertical-gap",
                        type=float, default=2.5,
                        help="""The vertical gap between labels.""")
    parser.add_argument("-v", "--verbose",
                        action='store_true',
                        help="""Produce explanatory output.""")
    parser.add_argument("-b", "--box",
                        action='store_true',
                        help="""Draw a box around each label.""")
    parser.add_argument("-o", "--output",
                        default="labels-%d.svg",
                        help="""The file to write the formatted labels into.""")
    parser.add_argument("input_file")
    args = parser.parse_args()

    columns = args.across
    rows = args.down

    left_margin = args.left_margin
    top_margin = args.top_margin
    right_margin = left_margin
    bottom_margin = top_margin
    label_width_with_gap = args.width + args.horizontal_gap
    label_height_with_gap = args.height + args.vertical_gap 
    page_width = left_margin + label_width_with_gap * columns + right_margin
    page_height = top_margin + label_height_with_gap * rows + bottom_margin

    config = {
        'outline': args.box,
        'leading': 1.2,
        'label-width': args.width,
        'label-height': args.height,
        'vertical-gap': args.vertical_gap,
        'horizontal-gap': args.horizontal_gap,
        'label-left-margin': 5,
        'label-top-margin': 0,
        'verbose': args.verbose
    }
    
    for ipage, page in enumerate(labels_to_pages(read_labels(args.input_file), rows * columns)):
        with open(args.output % ipage, 'w') as outstream:
            row = 0
            column = 0
            begin_page(outstream, page_width, page_height)
            for label in page:
                write_label_at(left_margin + column * label_width_with_gap,
                               top_margin + row * label_height_with_gap,
                               config,
                               label,
                               outstream)
                row += 1
                if row == rows:
                    if args.verbose:
                        print("next label below the last would be at", top_margin + row * label_height_with_gap)
                    row = 0
                    column += 1
            end_page(outstream)
    
if __name__ == "__main__":
    main()
    



#!/usr/bin/env python3

import argparse
import re
import coimealta.contacts.contacts_data as contacts_data
from collections import defaultdict

def offspring(person):
    return person['Offspring']

def parents(person):
    return person['Parents']

def partners(person):
    return person['Partners']

def siblings(person):
    return person['Siblings']

def name_to_ID(name, by_name):
    name = name.replace('_', ' ')
    if name not in by_name:
        print("Name", name, "not found")
    return by_name[name]['ID'] if name in by_name else "Z0Z0"

def normalize_to_IDs(people, by_name):
    return set([(person
                 if re.match("[G-Z][0-9][A-Z][0-9]", person)
                 else name_to_ID(person, by_name))
                for person in people])

def find_siblings(person, by_id):
    sibs = person['Siblings']
    more = True
    if len(sibs) > 0:
        while more:
            try:
                for sib in list(person['Siblings']):
                    sibsibs = by_id[sib]['Siblings']
                    more = False
                    for sibsib in sibsibs:
                        if sibsib not in sibs:
                            sibs.add(sibsib)
                            more = True
            except KeyError:
                print("Missing key in looking for sibling", sib, "of person", person)
                break
    yourself = person['ID']
    if yourself in sibs:
        sibs.remove(yourself)
    return sibs

def name(person):
    return person['_name_']

def accumulate(person, aspect, by_aspect):
    by_aspect[person[aspect]].append(person['ID'])

def print_summary(by_aspect, label):
    by_frequency = {}
    for k, v in by_aspect.items():
        frequency = len(v)
        if frequency not in by_frequency:
            by_frequency[frequency] = []
        by_frequency[frequency].append(k)
    print(label, "; ".join(
        [ "; ".join(
            ["%s(%d)" % (bf, freq)
             for bf in sorted(by_frequency[freq], key=lambda x: x if x else "")])
          for freq in reversed(sorted(by_frequency.keys()))]))

def analyze_contacts(by_id):
    by_nationality = defaultdict(list)
    by_gender = defaultdict(list)
    by_title = defaultdict(list)
    by_place_met = defaultdict(list)
    flagged = defaultdict(list)
    for id, person in by_id.items():
        accumulate(person, 'Nationality', by_nationality)
        accumulate(person, 'Gender', by_gender)
        accumulate(person, 'Title', by_title)
        accumulate(person, 'Place met', by_place_met)
        for flag in person.get('Flags', "") or "":
            flagged[flag].append(person['ID'])
    return {
        "n_people": len(by_id),
        "by_gender": by_gender,
        "by_nationality": by_nationality,
        "by_place_met": by_place_met,
        "by_title": by_title,
        "ordained": contacts_data.count_grouped_titles(by_title, ["Revd", "Revd Dr", "Revd Prof", "RtRevd"]),
        "doctored": contacts_data.count_grouped_titles(by_title, ["Dr", "Revd Dr", "Prof", "Revd Prof"]),
        "flagged": flagged
    }

def link_contacts(by_id, by_name):
    """Update the connections between people, in memory.
    Fills in the other direction for any that are given in only
    one direction."""

    for person in by_id.values():
        try:
            for field in ('Parents', 'Offspring', 'Siblings', 'Partners', 'Knows'):
                if person.get(field):
                    person[field] = normalize_to_IDs(person[field], by_name)
        except KeyError:
            print("missing key while processing", person)

    for person_id, person in by_id.items():
        partner_ids = person['Partners']
        if len(partner_ids) == 1: # don't try this on non-monogamists
            partner_id = next(iter(partner_ids))
            if partner_id not in by_id:
                print("Could not find partner", partner_id, "of", person['Given name'], person['Surname'])
                continue
            partner = by_id[partner_id]
            partners_partners = partner['Partners']
            if len(partners_partners) == 0: # again, for monogamists only
                partner['Partners'].add(person_id)
        for parent_id in person['Parents']:
            if parent_id not in by_id:
                print(person.get('_name_', person), "has an unlisted parent", parent_id)
                continue
            parent = by_id[parent_id]
            if person_id not in parent['Offspring']:
                parent['Offspring'].add(person_id)
        for offspring_id in person['Offspring']:
            if offspring_id not in by_id:
                print(person.get('_name_', person), "has an unlisted child", offspring_id)
                continue
            child = by_id[offspring_id]
            if person_id not in child['Parents']:
                child['Parents'].add(person_id)
        for sibling_id in find_siblings(person, by_id):
            if sibling_id not in by_id:
                print(person.get('_name_', person), "has an unlisted sibling", sibling_id)
                continue
            sibling = by_id[sibling_id]
            if person_id not in sibling['Siblings']:
                sibling['Siblings'].add(person_id)
        # todo: mutualize contacts

    return by_name

def link_contacts_main(input_file, analyze, graph, output_file):

    """Update the connections between people, in the contacts file.
    Fills in the other direction for any that are given in only
    one direction."""

    by_id, by_name = contacts_data.read_contacts(input_file)

    link_contacts(by_id, by_name)

    contacts_data.write_contacts(output_file, by_name)

    if graph:
        print("digraph {")
        for uid, person in by_id.items():
            their_partners = person['Partners']
            their_offspring = person['Offspring']
            their_parents = person['Parents']
            if len(their_partners) > 0 or len(their_offspring) > 0 or len(their_parents) > 0:
                print("  ", uid,
                      ('[label="' + person['_name_']
                       + '" shape=' + ("box" if person['Gender'] == 'm' else "diamond") + "]"))
            if their_partners:
                print("    ", uid, "->", "{", ",".join(their_partners), "}")
                print("    {rank=same", uid, " ".join(their_partners), "}")
            if their_offspring:
                print("    ", uid, "->", "{", ",".join(their_offspring), "} [style=dotted]")
                print("    {rank=same", " ".join(their_offspring), "}")
            if their_parents:
                print("    ", uid, "->", "{", ",".join(their_parents), "} [style=dashed]")
        print("}")

    if analyze:
        analysis = analyze_contacts(by_id)
        n_people = analysis['n_people']
        print(n_people, "people")
        print_summary(analysis['by_nationality'], "nationalities:")
        print_summary(analysis['by_gender'], "genders:")
        print_summary(analysis['by_title'], "titles:")
        print_summary(analysis['by_place_met'], "places met:")
        ordained = analysis['ordained']
        print("%d ordained (%d%% of the people you know)" % (ordained, ordained*100 / n_people))
        doctored = analysis['doctored']
        print("%d with doctorates (%d%% of the people you know)" % (doctored, doctored * 100 / n_people))
        return analysis
    else:
        return None

def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--analyze", action='store_true')
    parser.add_argument("--graph", action='store_true')
    parser.add_argument("input_file")
    parser.add_argument("output_file")
    return vars(parser.parse_args())

if __name__ == "__main__":
    link_contacts_main(**get_args())
